#include <avr/io.h>
#include <util/delay.h>

#define USART_BAUDRATE 125000
#define BAUD_PRESCALE (((F_CPU / (USART_BAUDRATE * 16UL))) - 1)

#define DDR_BM	DDRD
#define PORT_BM PORTD
#define PIN_BM PIND
#define TX_PIN 2
#define RX_PIN 3


static char *hex = "0123456789ABCDEF";


// Wait for RX to go low
void WaitRxLow()
{
	while (PIN_BM & (1 << RX_PIN));
}

// Wait for RX to go high, for at most 600ms
void WaitRxHigh()
{
	TIFR |= (1 << OCF1A);	// Clear the timer overflow flag, so we can listen for it again
	TCNT1 = 0;				// Set the counter to 0
	
	// While RX is low and we haven't reached the 600ms timeout
	while (	!(PIN_BM & (1 << RX_PIN)) && !(TIFR & (1 << OCF1A)) );
}

void serial_send(uint8_t b)
{
	while (!(UCSRA & (1 << UDRE)));
	UDR = b;
}

int main()
{
	// Set up USART
	UCSRB |= (1 << TXEN);   // Turn on transmission circuitry
	UCSRC |= (1 << UCSZ0) | (1 << UCSZ1);
		
	UBRRH = (BAUD_PRESCALE >> 8); // Load upper 8-bits of the baud rate value into the high byte of the UBRR register
	UBRRL = BAUD_PRESCALE; // Load lower 8-bits of the baud rate value into the low byte of the UBRR register


	// Set up the port
	PORT_BM &= ~(1 << RX_PIN);
	DDR_BM = (1 << TX_PIN);
	
	// Set up the timeout timer
	OCR1A = 19531;								// In combination with the 1/256 prescaller we'll get an overflow at roughly 600ms
	TCCR1A = 0;
	TCCR1B = (1 << WGM12) | (1 << CS12);		// CTC Mode and 1/256 prescaler
	
	uint8_t buff[20];
	
	while (1)
	{
		// Don't send the Tx signal if the meter is in the process of sending something
		WaitRxLow();
	
		// Send out the first signal
		PORT_BM |= (1 << TX_PIN);
		_delay_ms(10);
		PORT_BM &= ~(1 << TX_PIN);
		
		// Now wait for RX the to get high
		WaitRxHigh();
		
		// If Rx is high, let's pump some clock into it
		if ( PIN_BM & (1 << RX_PIN) )
		{
			// Read whatever many bytes we have a buffer for
			for (uint8_t i = 0; i < sizeof(buff); ++i)
			{
				uint8_t b = 0;
				
				for (uint8_t j = 0; j < 8; ++j)
				{
					// Rising edge / setup edge
					PORT_BM |= (1 << TX_PIN);
					_delay_us(75);
					
					
					// Do sampling here
					b >>= 1;
					b |= ((PIN_BM & (1 << RX_PIN)) >> (RX_PIN)) << 7;
					
					// Falling edge / sample edge
					PORT_BM &= ~(1 << TX_PIN);
					_delay_us(75);
				}
				
				buff[i] = ~b;
			}
			
			// Send whatever we've read out on the serial port
			for (uint8_t i = 0; i < sizeof(buff); ++i)
			{
				// Send first half of the byte
				serial_send(hex[buff[i] >> 4]);

				// Send the second half of the byte
				serial_send(hex[buff[i] & 0x0F]);
			}
			
			// Send out a new line as well
			serial_send('\r');
			serial_send('\n');
		}
		
		_delay_ms(20); 
	}

	return 0;
}
